#! /usr/bin/env Rscript
## This runs testme test script inst/testme/test-makeClusterSequential.R
## Don't edit - it was autogenerated by inst/testme/deploy.R
parallelly:::testme("makeClusterSequential")
