% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ports.R
\name{canPortBeUsed}
\alias{canPortBeUsed}
\title{Check whether a TCP port can be opened or not}
\usage{
canPortBeUsed(port)
}
\arguments{
\item{port}{(integer) A TCP port in [0, 65535].}
}
\value{
\code{canPortBeUsed(port)} returns a logical indicating whether the port can
be opened or not, or cannot be queried.  If the port can be opened,
then \code{TRUE} is returned, if cannot be opened then \code{FALSE} is returned,
which may happen if the port is used by another process.
If port querying is not supported, as in R (< 4.0.0),  then \code{NA} is
returned.
}
\description{
Check whether a TCP port can be opened or not
}
\keyword{internal}
