% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/serializedSize.R
\name{serializedSize}
\alias{serializedSize}
\title{Calculate the size of an R object when it is serialized}
\usage{
serializedSize(obj)
}
\arguments{
\item{obj}{An R object.}
}
\value{
(double) Number of bytes needed to serialize this object.
}
\description{
This function goes through all the motions of serializing an object, but
does nothing with the bytes other than to tally the total length.
}
\examples{
object.size(mtcars)
serializedSize(mtcars)

}
\author{
Mike FC
}
