% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/isForkedNode.R
\name{isForkedNode}
\alias{isForkedNode}
\title{Checks whether or not a Cluster Node Runs in a Forked Process}
\usage{
isForkedNode(node, ...)
}
\arguments{
\item{node}{A cluster node of class \code{SOCKnode} or \code{SOCK0node}.}

\item{\ldots}{Not used.}
}
\value{
(logical) Returns TRUE if the cluster node is running in a
forked child process and FALSE if it does not.
If it cannot be inferred, NA is returned.
}
\description{
Checks whether or not a Cluster Node Runs in a Forked Process
}
