% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/isLocalhostNode.R
\name{isLocalhostNode}
\alias{isLocalhostNode}
\title{Checks whether or not a Cluster Node Runs on Localhost}
\usage{
isLocalhostNode(node, ...)
}
\arguments{
\item{node}{A cluster node of class \code{SOCKnode} or \code{SOCK0node}.}

\item{\ldots}{Not used.}
}
\value{
(logical) Returns TRUE if the cluster node is running on the
current machine and FALSE if it runs on another machine.
If it cannot be inferred, NA is returned.
}
\description{
Checks whether or not a Cluster Node Runs on Localhost
}
