% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/isForkedChild.R
\name{isForkedChild}
\alias{isForkedChild}
\title{Checks whether or not we are running in a forked child process}
\usage{
isForkedChild()
}
\value{
(logical) Returns TRUE if the running in a forked child
process, otherwise FALSE.
}
\description{
Checks whether or not we are running in a forked child process
}
\details{
Examples of setups and functions that rely on \emph{forked} parallelization
are \code{parallel::makeCluster(n, type = "FORK")}, \code{parallel::mclapply()},
and \code{future::plan("multicore")}.
}
